﻿using System;
using System.Text;
using System.IO;
using System.Threading;
using System.Collections.Generic;

namespace SiteBuilder
{
    class Program
    {
        static string templateString;
        static string destFolder = @"bin";
        static string srcFolder = @"src";
        private static Timer m_timer;
        private static bool m_timerStarted = false;
        private static List<string> m_pendingEvents = new List<string>();

        private static bool Whatif = false;

        private static string TemplateFile
        {
            get { return Path.Combine(srcFolder, "template.html"); }
        }

        static void Main(string[] args)
        {
            // init stuff. make the file paths absolute, get the template string.
            srcFolder = Path.Combine(Environment.CurrentDirectory, srcFolder);
            destFolder = Path.Combine(Environment.CurrentDirectory, destFolder);
            ReloadTemplateFile();

            // do one round of compilation at startup
            //CompileHtml();
            // then keep watching for changes

            RecursiveCopy(srcFolder);
            BeginWatchingFileSystem();
            
        }

        private static void ReloadTemplateFile()
        {
            templateString = File.ReadAllText(TemplateFile);
        }

        private static void BeginWatchingFileSystem()
        {
            FileSystemWatcher fsw = new FileSystemWatcher();
            fsw.Path = srcFolder;
            fsw.NotifyFilter = NotifyFilters.LastWrite;
            fsw.IncludeSubdirectories = true;
            fsw.Filter = "*.*";
            fsw.Changed += fsw_Changed;
            m_timer = new Timer(OnTimeout, null, Timeout.Infinite, Timeout.Infinite);

            fsw.EnableRaisingEvents = true;
            // Wait for the user to quit the program.
            Console.WriteLine("Press \'q\' to quit the sample.");
            while (Console.Read() != 'q') ;
        }

        private static void OnTimeout(object state)
        {
            var localCopy = new List<string>();

            // Don't want other threads messing with the pending events right now
            lock (m_pendingEvents)
            {
                foreach (var item in m_pendingEvents)
                {
                    localCopy.Add(item);
                }

                // reset the pending events and the timer
                m_pendingEvents.Clear();
                m_timer.Change(Timeout.Infinite, Timeout.Infinite);
                m_timerStarted = false;
            }

            foreach (var item in localCopy)
            {
                Console.WriteLine(DateTime.Now.ToShortTimeString() + ": detected changes in: " + item);
                if (item == "template.html")
                {
                    ReloadTemplateFile();
                }
                RecursiveCopy(srcFolder);
            }
        }

        static void fsw_Changed(object sender, FileSystemEventArgs e)
        {
            lock (m_pendingEvents)
            {
                // Save a timestamp for the most recent event for this path
                if (!m_pendingEvents.Contains(e.Name))
                {
                    m_pendingEvents.Add(e.Name);
                }

                // Start a timer if not already started
                if (!m_timerStarted)
                {
                    m_timer.Change(100, 100);
                    m_timerStarted = true;
                }
            }
        }

        private static void RecursiveCopy(string currentFolder)
        {
            List<string> folders = new List<string>();

            // First, check the current folder for input files
            string[] files = Directory.GetFiles(currentFolder);
            foreach (var file in files)
            {
                switch (GetFileProcType(file))
                {
                    case FileProcType.HtmlFile:
                        CompileHtmlIfNeeded(file);
                        break;
                    case FileProcType.CssFile:
                        CompileCssIfNeeded(file);
                        break;
                    case FileProcType.ContentFile:
                        CopyFileIfNeeded(file);
                        break;
                    case FileProcType.TemplateFile:
                    case FileProcType.DontProcess:
                    default:
                        break;
                }
            }

            string[] dirs = Directory.GetDirectories(currentFolder);
            foreach (var dir in dirs)
            {
                RecursiveCopy(dir);
            }
        }

        private static void CopyFileIfNeeded(string src)
        {
            string dest = getOutputFilename(src);
            
            if (IsProcessingRequired(src, dest))
            {
                // create the folder if it does not yet exist
                CreateDirectoryIfNotExists(Path.GetDirectoryName(dest));
                if (Whatif) { Console.WriteLine("copying from {0} to {1}", src, dest); }
                else
                {
                    File.Copy(src, dest, true);
                }
            }
        }

        /// <summary>
        /// Creates a directory if it does not yet exist
        /// </summary>
        /// <param name="dir"></param>
        private static void CreateDirectoryIfNotExists(string dir)
        {
            if (!Directory.Exists(dir))
            {
                if (Whatif) { Console.WriteLine("mkdir: " + dir); }
                else
                {
                    Directory.CreateDirectory(dir);
                }
            }
        }

        /// <summary>
        /// Returns true if the destination file does not exist, or if it is older than the source file
        /// </summary>
        /// <param name="src">Source file</param>
        /// <param name="dest">destination file</param>
        /// <returns></returns>
        private static bool IsProcessingRequired(string src, string dest)
        {
            if (!File.Exists(dest)) return true;
            if (File.GetLastWriteTime(dest) < File.GetLastWriteTime(src)) return true;
            if (GetFileProcType(src) == FileProcType.HtmlFile)
            {
                if (File.GetLastWriteTime(dest) < File.GetLastWriteTime(TemplateFile)) return true;
            }
            return false;
        }

        private static void CompileHtmlIfNeeded(string src)
        {
            string dest = getOutputFilename(src);

            if (IsProcessingRequired(src, dest))
            {
                // create the folder if it does not yet exist
                CreateDirectoryIfNotExists(Path.GetDirectoryName(dest));

                // we assume template string is pre-populated
                string res = templateString;

                string[] lines = File.ReadAllLines(src);
                StringBuilder sb = new StringBuilder();
                for (int i = 0; i < lines.Length; i++)
                {
                    // if there is a variable here
                    if (lines[i].StartsWith("@ViewBag."))
                    {
                        var words = lines[i].Split('=');
                        res = res.Replace(words[0], words[1]); // replace in the document bag
                    }
                    else
                    {
                        sb.AppendLine(lines[i]);
                    }
                }

                if (Whatif) { Console.WriteLine("Writing to file: " + getOutputFilename(src)); }
                else
                {
                    File.WriteAllText(getOutputFilename(src),
                        res.Replace("@ViewBag.Document", sb.ToString()));
                }
            }
        }

        private static void CompileCssIfNeeded(string src)
        {
            string dest = getOutputFilename(src);

            if (IsProcessingRequired(src, dest))
            {
                string s = "";
                Dictionary<string, string> mappings = new Dictionary<string, string>();

                // start transposing input file to output file
                using (StreamReader rdr = new StreamReader(src))
                using (StreamWriter wrt = new StreamWriter(dest))
                while((s = rdr.ReadLine())!=null)
                {
                    if (s.StartsWith("$")) // there is a variable declaration here
                    {
                        var words = s.Split(':');
                        // map the variable name to the variable value
                        mappings[words[0].Trim()] = words[1].Trim();
                    }
                    else
                    {
                        foreach (var tok in GetCssTokens(s))
                        {
                            if (mappings.ContainsKey(tok))
                            {
                                wrt.Write(mappings[tok]);
                            }
                            else
                            {
                                wrt.Write(tok);
                            }
                        }
                        wrt.WriteLine();
                    }
                }
            }
        }

        /// <summary>
        /// Gets all the tokens in a css, split at variables that start with $.
        /// </summary>
        /// <param name="line">The input string</param>
        /// <returns>The input string, token by token, where the tokens of interest are variables that begin with the $ character.</returns>
        /// <remarks>Returns all tokens, not just the variable tokens</remarks>
        internal static IEnumerable<string> GetCssTokens(string line)
        {
            if (line.IndexOf('$') == -1) yield return line;
            else
            {
                // we have established that there are one or more tokens.
                int start = 0;
                int end = 0;
                bool isVariable = false;

                // iterate through the string
                foreach (char ch in line)
                {
                    if (!isVariable) // we are not in variable mode. iterate till we find the '$' sign
                    {
                        if (ch == '$') // we have found the beginning of a variable
                        {
                            if (end > start) // if there is a string to return, return it now
                            {
                                yield return line.Substring(start, end - start);
                            }
                            start = end; // mark this as the beginning of a new token
                            isVariable = true;
                        }
                    }
                    else // we are in variable mode. Look for any separator character
                    {
                        if (!(Char.IsLetterOrDigit(ch) || ch == '-'))
                        {
                            yield return line.Substring(start, end - start);
                            start = end; // mark this as the beginning of a new token
                            isVariable = (ch == '$'); // mark this as the beginning of a new variable if this char is a '$'
                        }
                    }

                    end++;
                }

                if (end > start) { yield return line.Substring(start, end - start); } // return the last token
            }
        }

        private static string getOutputFilename(string file)
        {
            switch (GetFileProcType(file))
            {
                case FileProcType.HtmlFile:
                {
                    var justTheName = Path.GetFileNameWithoutExtension(file);
                    var justTheFolder = Path.GetDirectoryName(file);
                    //var fname = new Uri(justTheFolder);
                    //var buri = new Uri(srcFolder);
                    //var fragment = buri.MakeRelativeUri(fname);
                    var fragment = GetRelativePath(justTheFolder, srcFolder);
                    string finalFragment = "";

                    if (justTheName == "index")
                    {
                        finalFragment = Path.Combine(fragment.ToString(), "index.html");
                    }
                    else
                    {
                        finalFragment = Path.Combine(fragment.ToString(), justTheName + "\\index.html");
                    }
                    
                    return Path.Combine(destFolder, finalFragment);
                }
                case FileProcType.ContentFile:
                case FileProcType.TemplateFile:
                case FileProcType.DontProcess:
                case FileProcType.CssFile:
                default:
                {
                    //var fname = new Uri(file);
                    //var buri = new Uri(srcFolder);
                    //var fragment = buri.MakeRelativeUri(fname);

                    return Path.Combine(destFolder, GetRelativePath(file, srcFolder));
                }
            }
        }

        private static string GetRelativePath(string fullFilename, string dirName)
        {
            if (fullFilename.StartsWith(dirName))
            {
                if (fullFilename == dirName) return "";
                return fullFilename.Substring(dirName.Length + 1);
            }

            throw new ArgumentException("aaaa");
        }

        private static FileProcType GetFileProcType(string filename)
        {
            if (Path.GetFileName(filename) == "template.html")
            {
                return FileProcType.TemplateFile;
            }
            else if (Path.GetExtension(filename) == ".html")
            {
                return FileProcType.HtmlFile;
            }
            else if (Path.GetExtension(filename) == ".css")
            {
                return FileProcType.CssFile;
            }
            else
            {
                return FileProcType.ContentFile;
            }
        }

        enum FileProcType
        {
            HtmlFile,
            ContentFile,
            TemplateFile,
            CssFile,
            DontProcess
        }
    }
}
